/********************************************/
/*  zalozeno na                             */
/*  SIXPACK.C -- Data compression program   */
/*  Written by Philip G. Gage, April 1991   */
/********************************************/

#include "sixpack.h"

#define TEXTSEARCH 1000   /* Max strings to search in text file */
#define BINSEARCH   200   /* Max strings to search in binary file */
#define TEXTNEXT     50   /* Max search at next character in text file */
#define BINNEXT      20   /* Max search at next character in binary file */
#define MAXFREQ    2000   /* Max frequency count before table reset */
#define MINCOPY       3   /* Shortest string copy length */
#define MAXCOPY      64   /* Longest string copy length */
#define SHORTRANGE    3   /* Max distance range for shortest length copy */
#define COPYRANGES    6   /* Number of string copy distance bit ranges */
short copybits[COPYRANGES] = {4,6,8,10,12,14};   /* Distance bits */

#define CODESPERRANGE (MAXCOPY - MINCOPY + 1)

#define NIL -1                    /* End of linked list marker */
#define HASHSIZE 16384            /* Number of entries in hash table */
#define HASHMASK (HASHSIZE - 1)   /* Mask for hash key wrap */

/* Adaptive Huffman variables */
#define TERMINATE 256             /* EOF code */
#define FIRSTCODE 257             /* First code for copy lengths */
#define SPMAXCHAR (FIRSTCODE+COPYRANGES*CODESPERRANGE-1)
#define SUCCMAX (SPMAXCHAR+1)
#define TWICEMAX (2*SPMAXCHAR+1)
#define ROOT 1
#define MAXBUF  4096    // musi byt vzdy sude cislo, maximalne 32766

int maxdistance;
int maxsize;


/*** Bit packing routines ***/


typedef struct Variables Variables,*P_Variables;

struct Variables {
    int copymin[COPYRANGES], copymax[COPYRANGES];
    int distance,insert,dictfile,binary;
    short *head;                                            /* Hash table */
    short *tail;
    short *succ;                                            /* Doubly linked lists */
    short *pred;
    unsigned char *Buffer;                                  /* Text buffer */
    unsigned short int Input_Bit_Count;                     /* Input bits buffered */
    unsigned short int Input_Bit_Buffer;                    /* Input buffer */
    unsigned short int Output_Bit_Count;                    /* Output bits buffered */
    unsigned short int Output_Bit_Buffer;                   /* Output buffer */
    long Bytes_In,Bytes_Out;                                /* File size counters */
    unsigned short int left[SPMAXCHAR+1],right[SPMAXCHAR+1];/* Huffman tree */
    unsigned short int up[TWICEMAX+1], freq[TWICEMAX+1];
    short int OutBufCount,InBufCount;
    unsigned short int  Buf16[MAXBUF];
    unsigned char       Buf8[MAXBUF];
    int input,output;
    int ArError;
    MD5_CTX MD5;
    P_local_files Local;
    uint8_t ItsDecode;
    int Readed,Writed;
};

void DoneVariables(P_Variables Vars)
{
    if(Vars==NULL)
        return;

    if(Vars->Buffer!=NULL)
    {
        FreeMem(Vars->Buffer);
        Vars->Buffer=NULL;
    }
    if(Vars->pred!=NULL)
    {
        FreeMem(Vars->pred);
        Vars->pred=NULL;
    }
    if(Vars->succ!=NULL)
    {
        FreeMem(Vars->succ);
        Vars->succ=NULL;
    }
    if(Vars->tail!=NULL)
    {
        FreeMem(Vars->tail);
        Vars->tail=NULL;
    }
    if(Vars->head!=NULL)
    {
        FreeMem(Vars->head);
        Vars->head=NULL;
    }
}

/* Initialize data for compression or decompression */
void Initialize(P_Variables Vars)
{
    short int i,j;

    /* Initialize Huffman frequency tree */
    for (i=2;i<=TWICEMAX;i++)
    {
        Vars->up[i]=i>>1;
        Vars->freq[i]=1;
    }
    for (i=1;i<=SPMAXCHAR;i++)
    {
        Vars->left[i]=i<<1;
        Vars->right[i]=(i<<1)+1;
    }
  /* Initialize copy distance ranges */
  j = 0;
  for (i = 0; i<COPYRANGES; i++) {
    Vars->copymin[i] = j;
    j += 1 << copybits[i];
    Vars->copymax[i] = j - 1;
  }
  maxdistance = j - 1;
  maxsize = maxdistance + MAXCOPY;
}

int InitVariables(P_local_files Local,P_Variables Vars)
{
    memset(Vars,0,sizeof(Variables));

    Initialize(Vars);

    Vars->insert=MINCOPY;

    Vars->head=(short int *)GetMem(HASHSIZE*(int)sizeof(short));
    Vars->tail=(short int *)GetMem(HASHSIZE*(int)sizeof(short));
    Vars->succ=(short int *)GetMem(maxsize*(int)sizeof(short));
    Vars->pred=(short int *)GetMem(maxsize*(int)sizeof(short));
    Vars->Buffer=(unsigned char *)GetMem(maxsize*(int)sizeof(unsigned char));

    if (Vars->head==NULL||Vars->tail==NULL||Vars->succ==NULL||Vars->pred==NULL||Vars->Buffer==NULL)
    {
        DoneVariables(Vars);
        Local->LastErrorCode=err_LowMemory;
        return 0;
    }

    memset(Vars->head,0,HASHSIZE*(int)sizeof(short));
    memset(Vars->tail,0,HASHSIZE*(int)sizeof(short));
    memset(Vars->succ,0,maxsize*(int)sizeof(short));
    memset(Vars->pred,0,maxsize*(int)sizeof(short));
    memset(Vars->Buffer,0,maxsize*(int)sizeof(unsigned char));

    return 1;
}

int WriteData(P_Variables Vars,void * Buffer,unsigned short OutSize)
{
    if(Vars->ArError!=0)
    {
        return 0;
    }
    if(OutSize>0)
    {
        if(Vars->output!=-1)
        {
            if(write(Vars->output,Buffer,OutSize)!=OutSize)
            {
                if(errno!=0)
                    Vars->ArError=errno;
                else
                    Vars->ArError=err_WriteError;
                return 0;
            }
        }
        if(Vars->ItsDecode)
            MD5Update(&Vars->MD5,(unsigned char *)Buffer,OutSize);
        Vars->Writed+=OutSize;
        Vars->OutBufCount=0;
    }
    return 1;
}

unsigned short ReadData(P_Variables Vars,void * Buffer,unsigned short InSize)
{
    if(Vars->ArError!=0)
    {
        return 0;
    }
    int Result=read(Vars->input,Buffer,InSize);
    if(Result!=InSize)
    {
        if(Result<0)
        {
            if(errno!=0)
            {
                Vars->ArError=errno;
            }
            else
            {
                Vars->ArError=err_ReadError;
            }
            return 0;
        }
    }
    if((Result)&&(!Vars->ItsDecode))
    {
        MD5Update(&Vars->MD5,(unsigned char *)Buffer,Result);
    }
    Vars->Readed+=Result;
    if(!Vars->ItsDecode)
    {
        time_t ActTime=time(NULL);
        Vars->Local->ActProgress+=Result;
        if(ActTime!=Vars->Local->ProgressTime)
        {
            Vars->Local->ProgressTime=ActTime;
            unsigned short Progress=0xFFFF;
            if(Vars->Local->TotalProgress!=0)
                Progress=(int)((double)Vars->Local->ActProgress/Vars->Local->TotalProgress*10000);
            SendWait(Vars->Local->client_sockfd,Vars->Local->ActCounter,Progress);
        }
    }
    return Result;
}

/* Write one bit to output file */
void Output_Bit(P_Variables Vars,short int bit)
{
    Vars->Output_Bit_Buffer<<=1;
    if(bit)
        Vars->Output_Bit_Buffer|=1;
    if(++Vars->Output_Bit_Count==16)
    {
        Vars->Buf16[Vars->OutBufCount]=Vars->Output_Bit_Buffer;
        Vars->Output_Bit_Count=0;
        Vars->Bytes_Out+=sizeof(unsigned short);
        if(++Vars->OutBufCount==MAXBUF)
        {
            WriteData(Vars,&Vars->Buf16,MAXBUF<<1);
        }
    }
}

/* Write multibit code to output file */
void Output_Code(P_Variables Vars,short int code,short int bits)
{
    short int i;
    for(i = 0; i<bits; i++)
    {
        Output_Bit(Vars,code & 0x01);
        code >>= 1;
    }
}

unsigned short Input_Code(P_Variables Vars,short int Bits)
{
    unsigned short i,code=0,Readed;
    unsigned short Bit=1;
    for(i=0;i<Bits;i++)
    {
        if(Vars->Input_Bit_Count==0)
        {
            if(Vars->InBufCount==MAXBUF)
            {
                Readed=ReadData(Vars,&Vars->Buf16,MAXBUF<<1);
                if(Readed==0)
                {
                    Vars->ArError=1;
                    return 0;
                }
                Vars->Bytes_In+=Readed;
                Vars->InBufCount=0;
            }
            Vars->Input_Bit_Buffer=Vars->Buf16[Vars->InBufCount++];
            Vars->Input_Bit_Count=15;
        }
        else
        {
            Vars->Input_Bit_Count--;
        }
        if((Vars->Input_Bit_Buffer&0x8000)!=0)
            code|=Bit;
        Vars->Input_Bit_Buffer<<=1;
        Bit<<=1;
    }
    return code;
}

/* Flush any remaining bits to output file before closing file */
void Flush_Bits(P_Variables Vars)
{
    if (Vars->Output_Bit_Count>0)
    {
        Vars->Output_Bit_Buffer<<=(16-Vars->Output_Bit_Count);
        Vars->Buf16[Vars->OutBufCount]=Vars->Output_Bit_Buffer;
        Vars->OutBufCount++;
        Vars->Output_Bit_Count=0;
        Vars->Bytes_Out+=sizeof(unsigned short);
    }
    WriteData(Vars,Vars->Buf16,Vars->OutBufCount<<1);
}

/*** Adaptive Huffman frequency compression ***/

/* Data structure based partly on "Application of Splay Trees
   to Data Compression", Communications of the ACM 8/88 */

/* Update frequency counts from leaf to root */
void Update_Freq(P_Variables Vars,short int a,short int b)
{
    do
    {
        Vars->freq[Vars->up[a]]=Vars->freq[a]+Vars->freq[b];
        a=Vars->up[a];
        if(a!=ROOT)
        {
            if(Vars->left[Vars->up[a]]==a)
                b=Vars->right[Vars->up[a]];
            else b=Vars->left[Vars->up[a]];
        }
    } while(a!=ROOT);

  /* Periodically scale frequencies down by half to avoid overflow */
  /* This also provides some local adaption and better compression */
    if(Vars->freq[ROOT]==MAXFREQ)
        for(a=1;a<=TWICEMAX;a++)
            Vars->freq[a]>>=1;
}

/* Update Huffman model for each character code */
void Update_Model(P_Variables Vars,unsigned short int code)
{
    short int a, b, c, ua, uua;
    a=code+SUCCMAX;
    ++Vars->freq[a];
    if(Vars->up[a]!=ROOT)
    {
        ua=Vars->up[a];
        if(Vars->left[ua]==a)
            Update_Freq(Vars,a,Vars->right[ua]);
        else Update_Freq(Vars,a,Vars->left[ua]);
        do
        {
            uua=Vars->up[ua];
            if(Vars->left[uua]==ua)
                b=Vars->right[uua];
            else b=Vars->left[uua];

            /* If high freq lower in tree, swap nodes */
            if(Vars->freq[a]>Vars->freq[b])
            {
                if(Vars->left[uua]==ua)
                    Vars->right[uua]=a;
                else Vars->left[uua]=a;
                if(Vars->left[ua]==a)
                {
                    Vars->left[ua]=b;c=Vars->right[ua];
                }
                else
                {
                    Vars->right[ua]=b;c=Vars->left[ua];
                }
                Vars->up[b]=ua;Vars->up[a]=uua;
                Update_Freq(Vars,b,c);a=b;
            }
            a=Vars->up[a];ua=Vars->up[a];
        } while (ua!=ROOT);
    }
}

/* Compress a character code to output stream */
void Compress(P_Variables Vars,unsigned short int code)
{
    short int a,sp=0;
    unsigned short int stack[50];

    a=code+SUCCMAX;
    do
    {
        stack[sp++]=(Vars->right[Vars->up[a]]==a);
        a=Vars->up[a];
    } while(a!=ROOT);
    do
    {
        Output_Bit(Vars,stack[--sp]);
    } while(sp);
    Update_Model(Vars,code);
}

/* Uncompress a character code from input stream */
int Uncompress(P_Variables Vars)
{
    unsigned short Readed;
    unsigned short a=1;
    do
    {
        if(Vars->Input_Bit_Count==0)
        {
            if(Vars->InBufCount==MAXBUF)
            {
                Readed=ReadData(Vars,&Vars->Buf16,MAXBUF<<1);
                if(Readed==0)
                {
                    Vars->ArError=1;
                    return 0;
                }
                Vars->InBufCount=0;
            }
            Vars->Input_Bit_Buffer=Vars->Buf16[Vars->InBufCount++];
            Vars->Input_Bit_Count=15;
        }
        else
        {
            Vars->Input_Bit_Count--;
        }
        if((Vars->Input_Bit_Buffer&0x8000)!=0)
            a=Vars->right[a];
        else
            a=Vars->left[a];
        Vars->Input_Bit_Buffer<<=1;
    } while(a<=SPMAXCHAR);
    a-=SUCCMAX;
    Update_Model(Vars,a);
    return a;
}

/*** Hash table linked list string search routines ***/

short int GetKey(P_Variables Vars,short int n)
{
    return ((Vars->Buffer[n]^(Vars->Buffer[(n+1)%maxsize]<<4)^(Vars->Buffer[(n+2)%maxsize]<<8)) & HASHMASK);
}

/* Add node to head of list */
void Add_Node(P_Variables Vars,short int n)
{

    short int key;

    key=GetKey(Vars,n);
    if(Vars->head[key]==NIL)
    {
        Vars->tail[key]=n;
        Vars->succ[n]=NIL;
    }
    else
    {
        Vars->succ[n]=Vars->head[key];
        Vars->pred[Vars->head[key]]=n;
    }
    Vars->head[key]=n;
    Vars->pred[n]=NIL;
}

/* Delete node from tail of list */
void Delete_Node(P_Variables Vars,short int n)
{

    short int key;

    key=GetKey(Vars,n);
    if(Vars->head[key]==Vars->tail[key])
        Vars->head[key]=NIL;
    else
    {
        Vars->succ[Vars->pred[Vars->tail[key]]]=NIL;
        Vars->tail[key]=Vars->pred[Vars->tail[key]];
    }
}

/* Find longest string matching lookahead buffer string */
int Match(P_Variables Vars,short int n,short int depth)
{
    short int i, j, index, key, dist, len, best = 0, count = 0;

    if(n==maxsize)
        n=0;
    key=GetKey(Vars,n);
    index=Vars->head[key];
    while(index!=NIL)
    {
        if(++count>depth)
            break;     /* Quit if depth exceeded */
        if(Vars->Buffer[(n+best)%maxsize]==Vars->Buffer[(index+best)%maxsize])
        {
            len=0;i=n;j=index;
            while(Vars->Buffer[i]==Vars->Buffer[j] && len<MAXCOPY && j!=n && i!=Vars->insert)
            {
                ++len;
                if(++i==maxsize)
                    i=0;
                if(++j==maxsize)
                    j=0;
            }
            dist=n-index;
            if(dist<0)
                dist+=maxsize;
            dist-=len;
            /* If dict file, quit at shortest distance range */
            if(Vars->dictfile&&dist>Vars->copymax[0])
                break;
            if(len>best&&dist<=maxdistance)
            {     /* Update best match */
                if(len>MINCOPY||dist<=Vars->copymax[SHORTRANGE+Vars->binary])
                {
                    best=len;Vars->distance=dist;
                }
            }
        }
        index=Vars->succ[index];
    }
    return(best);
}

/*** Finite Window compression routines ***/

#define IDLE 0    /* Not processing a copy */
#define COPY 1    /* Currently processing copy */

/* Check first buffer for ordered dictionary file */
/* Better compression using short distance copies */
void Dictionary(P_Variables Vars)
{
    short int i = 0, j = 0, k, count = 0;

    /* Count matching chars at start of adjacent lines */
    while (++j < MINCOPY+MAXCOPY)
    {
        if (Vars->Buffer[j-1]==10)
        {
            k=j;
            while(Vars->Buffer[i++]==Vars->Buffer[k++])
                ++count;
            i=j;
        }
    }
    /* If matching line prefixes > 25% assume dictionary */
    if(count>(MINCOPY+MAXCOPY)>>2)
        Vars->dictfile = 1;
}

int Complete(P_Variables Vars,char * MD5digest)
{
    Compress(Vars,TERMINATE);
    Flush_Bits(Vars);
    int RetInt;
    if(Vars->ArError==0)
        RetInt=1;
    else
    {
        RetInt=0;
        Vars->Local->LastErrorCode=Vars->ArError;
    }
    MD5Final((unsigned char *)MD5digest,&Vars->MD5);
    DoneVariables(Vars);
    FreeMem(Vars);
    return RetInt;
}

/* Encode file from input to output */
int Encode(P_local_files Local,int input,int output,char * MD5digest)
{
    short int c, i, n=MINCOPY, addpos=0, len=0, full=0, state=IDLE, nextlen, Result;

    P_Variables Vars=NULL;

    if((input<0)||(output<0))
    {
        Local->LastErrorCode=err_NotOpen;
        return 0;
    }
    if(MD5digest==NULL)
    {
        Local->LastErrorCode=err_BadParameters;
        return 0;
    }

    Vars=(Variables *)GetMem(sizeof(Variables));
    if(Vars==NULL)
    {
        Local->LastErrorCode=err_LowMemory;
        return 0;
    }

    if(!InitVariables(Local,Vars))
    {
        FreeMem(Vars);
        return 0;
    }

    Vars->input=input;
    Vars->output=output;
    Vars->Local=Local;

    MD5Init(&Vars->MD5);

    /* Initialize hash table to empty */
    for(i=0;i<HASHSIZE;i++)
    {
        Vars->head[i] = NIL;
    }

    Result=ReadData(Vars,&Vars->Buf8,MAXBUF);

    if(Vars->ArError==0)
    {
        /* Compress first few characters using Huffman */
        for(i=0;i<MINCOPY;i++)
        {
            if(Vars->InBufCount>=Result)
            {
                return Complete(Vars,MD5digest);
            }
            c=Vars->Buf8[Vars->InBufCount++];
            Compress(Vars,c);
            Vars->Bytes_In++;
            Vars->Buffer[i]=c;
            if(Vars->InBufCount==Result)
            {
                return Complete(Vars,MD5digest);
            }

        }
        /* Preload next few characters into lookahead buffer */
        for(i=0;i<MAXCOPY;i++)
        {
            c=Vars->Buf8[Vars->InBufCount++];
            Vars->Buffer[Vars->insert++]=c;
            Vars->Bytes_In++;
            if(c>127)
                Vars->binary=1;       /* Binary file ? */
            if(Vars->InBufCount==Result)
                break;
        }
        Dictionary(Vars);  /* Check for dictionary file */

        while((n!=Vars->insert)&&(Vars->ArError==0))
        {
            /* Check compression to insure really a dictionary file */
            if (Vars->dictfile && ((Vars->Bytes_In % MAXCOPY) == 0))
                if (Vars->Bytes_In/Vars->Bytes_Out < 2)
                    Vars->dictfile = 0;     /* Oops, not a dictionary file ! */

            /* Update nodes in hash table lists */
            if (full)
                Delete_Node(Vars,Vars->insert);
            Add_Node(Vars,addpos);

            /* If doing copy, process character, else check for new copy */
            if (state==COPY)
            {
                if (--len==1)
                    state=IDLE;
            }
            else
            {
                /* Get match length at next character and current char */
                if (Vars->binary)
                {
                    nextlen=Match(Vars,n+1,BINNEXT);
                    len=Match(Vars,n,BINSEARCH);
                }
                else
                {
                    nextlen=Match(Vars,n+1,TEXTNEXT);
                    len=Match(Vars,n,TEXTSEARCH);
                }

                /* If long enough and no better match at next char, start copy */
                if (len >= MINCOPY && len >= nextlen)
                {
                    state = COPY;

                    /* Look up minimum bits to encode distance */
                    for(i=0;i<COPYRANGES;i++)
                    {
                        if(Vars->distance<=Vars->copymax[i])
                        {
                            Compress(Vars,FIRSTCODE-MINCOPY+len+i*CODESPERRANGE);
                            Output_Code(Vars,Vars->distance-Vars->copymin[i],copybits[i]);
                            break;
                        }
                    }
                }
                else   /* Else output single literal character */
                    Compress(Vars,Vars->Buffer[n]);
            }

            /* Advance buffer pointers */
            if(++n==maxsize)
                n=0;
            if(++addpos==maxsize)
                addpos = 0;

            /* Add next input character to buffer */
            if(Vars->InBufCount<Result)
            {
                c=Vars->Buf8[Vars->InBufCount++];
                if(Vars->InBufCount==MAXBUF)
                {
                    Result=ReadData(Vars,&Vars->Buf8,MAXBUF);
                    Vars->InBufCount=0;
                }
                Vars->Buffer[Vars->insert++]=c;
                Vars->Bytes_In++;
                if(Vars->insert==maxsize)
                {
                    Vars->insert=0;
                    full=1;
                }
            }
            else
                full=0;
        }

    }
    return Complete(Vars,MD5digest);
}


/* Decode file from input to output */
int Decode(P_local_files Local,int input,int output,char * MD5digest)
{
    int c, i, j, k, dist, len, n = 0, index;

    P_Variables Vars=NULL;

    if(input<0)
    {
        Local->LastErrorCode=err_NotOpen;
        return 0;
    }

    if(MD5digest==NULL)
    {
        Local->LastErrorCode=err_BadParameters;
        return 0;
    }

    Vars=(Variables *)GetMem(sizeof(Variables));
    if(Vars==NULL)
    {
        Local->LastErrorCode=err_LowMemory;
        return 0;
    }

    if(!InitVariables(Local,Vars))
    {
        FreeMem(Vars);
        return 0;
    }

    Vars->input=input;
    Vars->output=output;
    Vars->Local=Local;
    Vars->ItsDecode=1;

    MD5Init(&Vars->MD5);

    Vars->InBufCount=MAXBUF;

    while(((c=Uncompress(Vars))!=TERMINATE)&&(Vars->ArError==0))
    {
        if(c<256)
        {     /* Single literal character ? */
            Vars->Buf8[Vars->OutBufCount++]=c;
            if(Vars->OutBufCount==MAXBUF)
            {
                WriteData(Vars,&Vars->Buf8,MAXBUF);
                Vars->Bytes_Out+=MAXBUF;
            }
            Vars->Buffer[n++]=c;
            if(n==maxsize)
                n=0;
        }
        else
        {            /* Else string copy length/distance codes */
            index=(c-FIRSTCODE)/CODESPERRANGE;
            len=c-FIRSTCODE+MINCOPY-index*CODESPERRANGE;
            dist=Input_Code(Vars,copybits[index])+len+Vars->copymin[index];
            j=n;k=n-dist;
            if(k<0)
                k+=maxsize;
            for(i=0;i<len;i++)
            {
                Vars->Buf8[Vars->OutBufCount++]=Vars->Buffer[k];
                if(Vars->OutBufCount==MAXBUF)
                {
                    WriteData(Vars,&Vars->Buf8,MAXBUF);
                    Vars->Bytes_Out+=MAXBUF;
                }
                Vars->Buffer[j++]=Vars->Buffer[k++];
                if(j==maxsize)
                    j=0;
                if(k==maxsize)
                    k=0;
            }
            n+=len;
            if(n>=maxsize)
                n-=maxsize;
        }
    }
    if(Vars->OutBufCount)
    {
        WriteData(Vars,&Vars->Buf8,Vars->OutBufCount);
        Vars->Bytes_Out+=Vars->OutBufCount;
    }
    i=Vars->ArError==0;
    if(!i)
    {
        SetLastFileError(Local,Vars->ArError,err_CorruptedFile);
    }
    MD5Final((unsigned char *)MD5digest,&Vars->MD5);
    DoneVariables(Vars);
    FreeMem(Vars);
    return i;
}

